/*
 * Author: Dušan Kolář
 */

#include "parse_dlms.h"
#include "dlms_data.h"
#include "vlq_base128_be.h"

#include <stddef.h>

#define SWAP_TYPE(WIDTH) typedef union u_swap ## WIDTH {\
    uint ## WIDTH ## _t value;\
    uint8_t  bytes[WIDTH/8];\
} swap ## WIDTH ## _t;

SWAP_TYPE(64)
SWAP_TYPE(32)
SWAP_TYPE(16)


typedef union u_invoke_id_and_priority {
    uint8_t    priority:1, service_class:1, reserved:2, invoke_id:4;
    uint8_t    value;
} invoke_id_and_priority_t;

// .................................................................
// prototypes

unsigned get_data_access_result_list(uint8_t **input_ptr, uint32_t* size, hdlc_parse_error_t *error);
unsigned get_data_access_result(uint8_t **input_ptr, uint32_t* size, hdlc_parse_error_t *error);
unsigned get_date_time_opt(uint8_t **input_ptr, uint32_t* size, hdlc_parse_error_t *error);
uint16_t get_length_encoded_num(uint8_t **input_ptr, uint32_t* size, hdlc_parse_error_t *error);
unsigned get_dlms_data_type(uint8_t *input, uint32_t size, hdlc_parse_error_t *error);
unsigned get_dlms_data_type_opt(uint8_t *input, uint32_t size, hdlc_parse_error_t *error);
void get_cosem_object_instance_id(uint8_t **input_ptr);
void get_invoke_id_and_priority(uint8_t value);

dlms_pdu_t *dlms_action_response(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);
dlms_pdu_t *dlms_set_response(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);
dlms_pdu_t *dlms_get_response(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);
dlms_pdu_t *dlms_action_request(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);
dlms_pdu_t *dlms_notification_request(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);
dlms_pdu_t *dlms_set_request(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);
dlms_pdu_t *dlms_get_request(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);

dlms_pdu_t *parse_dlms(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);

// .................................................................
// .................................................................

static dlms_pdu_t result;


dlms_pdu_t *parse_dlms(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error) {
    uint8_t *input = *input_ptr;

    if (size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    uint8_t pdu_type = *input++;
    --size;
    *input_ptr = input;
    
    result.pdu_type = pdu_type;
    result.data_block = NULL;
    result.req.data_present = 0;
    result.req.data_hard_present = 0;
    result.item_count = 0;
    result.date_time_present = 0;
    result.date_time = NULL;
    result.str = NULL;
    result.str_length = 0;
    result.list_length = 0;
    
    switch (pdu_type) {
        case
            DLMS_PDU_GET_REQUEST:
            return dlms_get_request(input_ptr,size,error);
        
        case
            DLMS_PDU_SET_REQUEST:
            return dlms_set_request(input_ptr,size,error);
        
        case
            DLMS_PDU_NOTIFICATION_REQUEST:
            return dlms_notification_request(input_ptr,size,error);
        
        case
            DLMS_PDU_ACTION_REQUEST:
            return dlms_action_request(input_ptr,size,error);
        
        case
            DLMS_PDU_GET_RESPONSE:
            return dlms_get_response(input_ptr,size,error);

        case
            DLMS_PDU_SET_RESPONSE:
            return dlms_set_response(input_ptr,size,error);

        case
            DLMS_PDU_ACTION_RESPONSE:
            return dlms_action_response(input_ptr,size,error);
        
        default:
            *error = HDLC_DLMS_WRONG_PDU_TYPE;
            return (NULL);
    } // switch
    
}

// .................................................................
// .................................................................

dlms_pdu_t *dlms_get_request(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error) {
    uint8_t *input = *input_ptr;

    if (size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    uint8_t request_type = *input++;
    --size;
    result.pdu_sub_type = request_type;
    *error = HDLC_OK;
    
    switch (request_type) {
        case
            DLMS_GET_REQ_NORMAL:
            
            if (size<11) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            get_cosem_object_instance_id(&input);
            size -= 10;
            if (get_dlms_data_type_opt(input, size, error)) {
                return (NULL);
            }
            return (&result);
            
        case
            DLMS_GET_REQ_NEXT:

            if (size<5) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            
            swap32_t block_num_raw;
            block_num_raw.bytes[3] = *input++;
            block_num_raw.bytes[2] = *input++;
            block_num_raw.bytes[1] = *input++;
            block_num_raw.bytes[0] = *input++;
            
            result.block_number = block_num_raw.value;
            
            return (&result);
            
        case
            DLMS_GET_REQ_WITH_LIST:
            
            if (size<11) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            get_cosem_object_instance_id(&input);
            size -= 10;
            if (get_dlms_data_type_opt(input, size, error)) {
                return (NULL);
            }
            return (&result);
        
        default:
            *error = HDLC_DLMS_WRONG_GET_REQ_TYPE;
            return (NULL);
    } // switch

}

// .................................................................

dlms_pdu_t *dlms_set_request(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error) {
    uint8_t *input = *input_ptr;

    if (size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    uint8_t request_type = *input++;
    --size;
    result.pdu_sub_type = request_type;
    *error = HDLC_OK;
    
    switch (request_type) {
        case
            DLMS_SET_REQ_NORMAL:
            
            if (size<11) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            get_cosem_object_instance_id(&input);
            size -= 10;
            if (get_dlms_data_type_opt(input, size, error)) {
                return (NULL);
            }
            
            if (result.req.data_present) {
                return (&result);
            }
            --size;
            ++input;
            
            // hard data
            result.req.data_hard_present = 1;
            if (get_dlms_data_type(input, size, error)) {
                return (NULL);
            }
            
            return (&result);
            
        case
            DLMS_SET_REQ_WITH_1_DBLCK:
            
            if (size<2) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);            
            result.data_block = input;
            
            return (&result);

        case
            DLMS_SET_REQ_WITH_DBLCK:

            if (size<11) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            get_cosem_object_instance_id(&input);
            size -= 10;
            if (get_dlms_data_type_opt(input, size, error)) {
                return (NULL);
            }
            
            if (result.req.data_present) {
                return (&result);
            }
            --size;
            ++input;
            
            // data block
            result.data_block = input;
            return (&result);
            
        case
            DLMS_SET_REQ_WITH_LIST:
            
            if (size<2) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            result.item_count = get_length_encoded_num(&input, &size, error);
            
            // Sequence not parsed, no hard data parsed
            
            if (*error == HDLC_OK) {
                return (&result);
            }
            return (NULL);
            
        break;
        case
            DLMS_SET_REQ_W_LIST_1_DBLCK:
            
            if (size<2) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            result.item_count = get_length_encoded_num(&input, &size, error);
            
            // Sequence not parsed, no BLOCK identified
            
            if (*error == HDLC_OK) {
                return (&result);
            }
            return (NULL);
    
        default:
            *error = HDLC_DLMS_WRONG_SET_REQ_TYPE;
            return (NULL);
    }
    
}

// .................................................................

dlms_pdu_t *dlms_notification_request(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error) {

    if (size<11) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    if (get_date_time_opt(input_ptr, &size, error)) {
        // Error during date_time parsing
        return (NULL);
    }

    if (size<10) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    get_cosem_object_instance_id(input_ptr);
    size -= 9;
    
    result.req.data_hard_present = 1;
    if (get_dlms_data_type(*input_ptr, size, error)) {
        return (NULL);
    }
            
    return (&result);
}

// .................................................................

dlms_pdu_t *dlms_action_request(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error) {
    uint8_t *input = *input_ptr;

    if (size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    uint8_t request_type = *input++;
    --size;
    result.pdu_sub_type = request_type;
    *error = HDLC_OK;
    
    switch (request_type) {
        case
            DLMS_ACT_REQ_NORMAL:
            
            if (size<11) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            get_cosem_object_instance_id(&input);
            size -= 10;
            if (get_dlms_data_type_opt(input, size, error)) {
                return (NULL);
            }
            return (&result);
        
        case
            DLMS_ACT_REQ_NEXT_PBLOCK:
            
            if (size<5) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            
            swap32_t block_num_raw;
            block_num_raw.bytes[3] = *input++;
            block_num_raw.bytes[2] = *input++;
            block_num_raw.bytes[1] = *input++;
            block_num_raw.bytes[0] = *input++;
            
            result.block_number = block_num_raw.value;
            
            return (&result);
        
        case
            DLMS_ACT_REQ_LIST:
        case
            DLMS_ACT_REQ_1PBLOCK:
        case
            DLMS_ACT_REQ_LIST_1PBLOCK:
        case
            DLMS_ACT_REQ_PBLOCK:
            
            // no data analysis done

            return (&result);

        default:
            *error = HDLC_DLMS_WRONG_ACT_REQ_TYPE;
            return (NULL);
    } // switch
    
}

// .................................................................

dlms_pdu_t *dlms_get_response(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error) {
    uint8_t *input = *input_ptr;

    if (size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    uint8_t response_type = *input++;
    --size;
    result.pdu_sub_type = response_type;
    *error = HDLC_OK;
    
    switch (response_type) {
        case
            DLMS_GET_RESP_NORMAL:
            
            if (size<2) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            --size;
            
            // hard data
            result.req.data_hard_present = 1;
            if (get_dlms_data_type(input, size, error)) {
                return (NULL);
            }
            
            return (&result);
            
        case
            DLMS_GET_RESP_NEXT:
            
            if (size<8) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            --size;
            if (get_data_access_result(&input, &size, error)) {
                return (NULL);
            }
            
            return (&result);
            
        case
            DLMS_GET_RESP_LIST:
            
            if (size<2) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            --size;
        
            if (get_data_access_result_list(&input, &size, error)) {
                return (NULL);
            }
            
            return (&result);
            
        default:
            *error = HDLC_DLMS_WRONG_GET_RESP_TYPE;
            return (NULL);
    } // switch
}

// .................................................................

dlms_pdu_t *dlms_set_response(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error) {
    uint8_t *input = *input_ptr;

    if (size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    uint8_t response_type = *input++;
    --size;
    result.pdu_sub_type = response_type;
    *error = HDLC_OK;
    
    switch (response_type) {
        case
            DLMS_SET_RESP_NORMAL:
            
            if (size!=2) {
                *error = HDLC_DLMS_SIZE_MISMATCH;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            
            result.data_access_result = *input;
            
            return (&result);
            
        case
            DLMS_SET_RESP_DBLCK:
            
            if (size!=5) {
                *error = HDLC_DLMS_SIZE_MISMATCH;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            
            swap32_t block_num_raw;
            block_num_raw.bytes[3] = *input++;
            block_num_raw.bytes[2] = *input++;
            block_num_raw.bytes[1] = *input++;
            block_num_raw.bytes[0] = *input++;
            
            result.block_number = block_num_raw.value;
            
            return (&result);            
            
        case
            DLMS_SET_RESP_LAST_DBLCK:
            
            if (size!=6) {
                *error = HDLC_DLMS_SIZE_MISMATCH;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            
            result.data_access_result = *input++;
            
            block_num_raw.bytes[3] = *input++;
            block_num_raw.bytes[2] = *input++;
            block_num_raw.bytes[1] = *input++;
            block_num_raw.bytes[0] = *input++;
            
            result.block_number = block_num_raw.value;
            
            return (&result);            
            
        case
            DLMS_SET_RESP_LAST_DBLCK_LIST:
            
            if (size<6) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            --size;
            
            if (get_data_access_result_list(&input, &size, error)) {
                return (NULL);
            }
            
            if (size!=4) {
                *error = HDLC_DLMS_SIZE_MISMATCH;
                return (NULL);
            }

            block_num_raw.bytes[3] = *input++;
            block_num_raw.bytes[2] = *input++;
            block_num_raw.bytes[1] = *input++;
            block_num_raw.bytes[0] = *input++;
            
            result.block_number = block_num_raw.value;
            
            return (&result);            
            
        case
            DLMS_SET_RESP_LIST:
            
            if (size<2) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            --size;
            
            if (get_data_access_result_list(&input, &size, error)) {
                return (NULL);
            }
            
            return (&result);            
            
            
        default:
            *error = HDLC_DLMS_WRONG_SET_RESP_TYPE;
            return (NULL);
    } // switch
}

// .................................................................

dlms_pdu_t *dlms_action_response(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error) {
    uint8_t *input = *input_ptr;

    if (size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (NULL);
    }
    
    uint8_t response_type = *input++;
    --size;
    result.pdu_sub_type = response_type;
    *error = HDLC_OK;
    
    switch (response_type) {
        case
            DLMS_ACT_RESP_NORMAL:
            
            if (size<2) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (NULL);
            }
            get_invoke_id_and_priority(*input++);
            --size;
            
            result.action_response_result = *input++;
            --size;
            
            if (get_dlms_data_type_opt(input, size, error)) {
                return (NULL);
            }
            
            if (result.req.data_present) {
                return (&result);
            }
            --size;
            ++input;

            result.data_access_result = *input;
            
            return (&result);
            
            
        case
            DLMS_ACT_RESP_PBLOCK:
        case
            DLMS_ACT_RESP_LIST:
        case
            DLMS_ACT_RESP_NEXT_PBLOCK:

            // no data analysis done

            return (&result);


        default:
            *error = HDLC_DLMS_WRONG_ACT_RESP_TYPE;
            return (NULL);
    }
}


// .................................................................
// .................................................................


void get_invoke_id_and_priority(uint8_t value) {
    invoke_id_and_priority_t iiap;
    iiap.value = value;
    result.priority = iiap.priority;
    result.service_class = iiap.service_class;
    result.invoke_id = iiap.invoke_id;
}

// .................................................................

void get_cosem_object_instance_id(uint8_t **input_ptr) {
    uint8_t *input = *input_ptr;
    
    swap16_t class_id_raw;
    class_id_raw.bytes[1] = *input++;
    class_id_raw.bytes[0] = *input++;
    result.req.class_id = class_id_raw.value;
    
    result.req.instance_id = input;
    input += 6;
    
    result.req.attribute_method_id = *input++;
    
    *input_ptr = input;
}

// .................................................................

unsigned get_dlms_data_type_opt(uint8_t *input, uint32_t size, hdlc_parse_error_t *error) {
    result.req.data_present = *input++;
    
    if (!result.req.data_present) return (0);  // no data

    if (--size<2) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (1);
    }
    
    result.req.access_selector = *input++;
    --size;
    
    return (get_dlms_data_type(input,size,error));
}

// .................................................................

unsigned get_dlms_data_type(uint8_t *input, uint32_t size, hdlc_parse_error_t *error) {    
    if (size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (1);
    }
    
    result.req.data_result_type = *input++;
    --size;
    
    if (result.req.data_result_type == 0) {
        
        if (size<1) {
            *error = HDLC_DLMS_TOO_SHORT;
            return (1);
        }
        result.req.data_type = *input++;
       --size;
    
        if (test_dlms_data_type(result.req.data_type)<0) {
            *error = HDLC_DLMS_WRONG_DATA_TYPE;
            return (1);
        }
        
        if (result.req.data_type == DLMS_OCTET_STRING) {
            if (size<2) {
                *error = HDLC_DLMS_TOO_SHORT;
                return (1);
            }
            result.req.octet_string_len = *input++;
            result.req.octet_string_start = input;
        }
        
    } else {  // should be one only
        
        if (size<1) {
            *error = HDLC_DLMS_TOO_SHORT;
            return (1);
        }
        result.data_access_result = *input++;
       --size;
        
    }
    
    return (0);
}

// .................................................................

uint16_t get_length_encoded_num(uint8_t **input_ptr, uint32_t* size, hdlc_parse_error_t *error) {
    if (*size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (0);
    }
    uint8_t *input = *input_ptr;
    uint8_t minSize = *input++;
    *size -= 1;
    if ((minSize & 0x80) == 0) {
        *input_ptr = input;
        *error = HDLC_OK;
        return ((uint16_t)minSize);
    }

    if (minSize != 0x82) {
        *error = HDLC_DLMS_UNSUPPORTED_LEN;
        return (0);
    }
    if (*size<2) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (0);
    }

    swap16_t res_raw;
    res_raw.bytes[1] = *input++;
    res_raw.bytes[0] = *input++;    
    *size -= 2;
    *input_ptr = input;
    
    *error = HDLC_OK;
    return (res_raw.value);
}

// .................................................................

unsigned get_date_time_opt(uint8_t **input_ptr, uint32_t* size, hdlc_parse_error_t *error) {
    if (*size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (1);  // ERROR
    }
    uint8_t *input = *input_ptr;
    result.date_time_present = *input++;
    *size -= 1;
    
    if (!result.date_time_present) {
        *input_ptr = input;
        *error = HDLC_OK;
        return (0); // OK
    }
    
    if (*size<12) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (1); // ERROR
    }
    
    result.date_time = input;
    input += 12;
    *size -= 1;
    
    *input_ptr = input;
    *error = HDLC_OK;
    return (0); // OK    
}

// .................................................................

unsigned get_data_access_result(uint8_t **input_ptr, uint32_t* size, hdlc_parse_error_t *error) {
    if (*size<7) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (1); // ERROR
    }
    uint8_t *input = *input_ptr;
    
    result.last_block = *input++;
    *size -= 1;
    
    swap32_t block_num_raw;
    block_num_raw.bytes[3] = *input++;
    block_num_raw.bytes[2] = *input++;
    block_num_raw.bytes[1] = *input++;
    block_num_raw.bytes[0] = *input++;
    *size -= 4;
    
    result.block_number = block_num_raw.value;
    
    result.result_choice = *input++;
    *size -= 1;
    
    switch (result.result_choice) {
        case 0:
            result.str_length = get_length_encoded_num(&input, size, error);
            if (*error != HDLC_OK) {
                return (1); // ERROR
            }
            result.str = (char*)input;
            break;
            
        case 1:
            result.data_access_result = *input++;
            *size -= 1;
            break;
            
        default:
            *error = HDLC_DLMS_WRONG_ACC_RES_CHOICE;
            return (1); // ERROR
    }
    
    *input_ptr = input;
    
    return (0); // OK
}


// .................................................................

unsigned get_data_access_result_list(uint8_t **input_ptr, uint32_t* size, hdlc_parse_error_t *error) {
    if (*size<1) {
        *error = HDLC_DLMS_TOO_SHORT;
        return (1); // ERROR
    }
    result.list_length = get_length_encoded_num(input_ptr, size, error);
    *size -= 1;
    if (result.list_length > MAX_LIST_LENGTH) {
        *error = HDLC_OUT_OF_MEMORY;
        return (1); // ERROR
    }
    if (result.list_length != *size) {
        *error = HDLC_DLMS_SIZE_MISMATCH;
        return (1); // ERROR
    }
    
    uint8_t *input = *input_ptr;
    unsigned num_of_res = (unsigned)result.list_length;
    *size -= num_of_res;
    for (unsigned u=0; u<num_of_res; ++u) {
        result.list_data_acces_result[u] = *input++;
    }
    
    *input_ptr = input;
    
    return (0); // OK
}


// EOF parse_dlms
// .................................................................
// .................................................................
