/*
 * Author: Dušan Kolář
 */

#include "parse_dlms.h"
#include "vlq_base128_be.h"

#include <stddef.h>

#define MIN_PACKET_SIZE    9
#define MAGIC              0x7E
#define LLC_MAGIC          0xE6
#define LLC_QUALITY        0


#define SWAP_TYPE(WIDTH) typedef union u_swap ## WIDTH {\
    uint ## WIDTH ## _t value;\
    uint8_t  bytes[WIDTH/8];\
} swap ## WIDTH ## _t;

SWAP_TYPE(64)
SWAP_TYPE(32)
SWAP_TYPE(16)

// .................................................................
// Prototypes

static uint8_t parse_llc(uint8_t **input, hdlc_parse_error_t *error);


// .................................................................
// .................................................................


typedef struct s_hdlc_format {
    uint16_t length:11,
             segmentation_flag:1,
             frame_format_type:4;
} hdlc_format_t;

typedef struct s_i_frame {
    uint8_t  zero_bit:1,
             send_sequence_number:3,
             pf_bit:1,                  // 0: hdlc_poll, 1: hdlc_final
             recv_sequence_number:3;
} i_frame_t;

typedef struct s_s_frame {
    uint8_t  one_bit:1,
             zero_bit:1,
             type:2,
             pf_bit:1,                  // 0: hdlc_poll, 1: hdlc_final
             recv_sequence_number:3;
} s_frame_t;

typedef struct s_u_frame {
    uint8_t  one_bit2:1,
             one_bit1:1,
             s_type:2,
             pf_bit:1,                  // 0: hdlc_poll, 1: hdlc_final
             r_type:3;
} u_frame_t;

typedef union u_ctrl_flags {
    uint8_t   value;
    i_frame_t i_frame;
    s_frame_t s_frame;
    u_frame_t u_frame;
} ctrl_flags_t;

// .................................................................

dlms_pdu_t *parse_hdlc(uint8_t *input, uint32_t size, hdlc_parse_error_t *error) {
    // test if the packet is sufficiently long
    if (size < MIN_PACKET_SIZE) {
        *error = HDLC_TOO_SHORT;
        return (NULL);
    }
    
    // test magic byte
    if (*input != MAGIC) {
        *error = HDLC_WRONG_MAGIC;
        return (NULL);
    }
    ++input;
    
    // read format field
    swap16_t format_raw;
    format_raw.bytes[1] = *input++;
    format_raw.bytes[0] = *input++;
    
    // transform BE format field do LE and bit-field structure
    hdlc_format_t format = *(hdlc_format_t*)((void*)&(format_raw.value));
    if (format.length != (size-2)) { // magic on the beginning and on the end doesn't counts
        *error = HDLC_NOT_MATCHING_SIZE;
        return (NULL);
    }
    size -= 3;
    
    uint64_t dst_address, src_address;
    int int_seq_err = 0; 
    
    // read destination address as BASE128 BER encoding
    dst_address = readSeq(&input,&size,&int_seq_err);
    if (int_seq_err != VLQ_ERR_OK) {
        *error = HDLC_BASE_SEQ;
        return (NULL);
    }
    
    // read source address as BASE128 BER encoding
    src_address = readSeq(&input,&size,&int_seq_err);
    if (int_seq_err != VLQ_ERR_OK) {
        *error = HDLC_BASE_SEQ;
        return (NULL);
    }
    
    if (size == 0) {
        *error = HDLC_OUT_OF_DATA;
        return (NULL);
    }
    
    // read 1 byte control flags
    ctrl_flags_t ctrl_flags;
    ctrl_flags.value = *input++;
    --size;
    
    if (ctrl_flags.i_frame.zero_bit != 0) {
        *error = HDLC_NOT_AN_IFRAME;
        return (NULL);
    }
    
    uint16_t hcs = 0;
    // read HCS 2byte word if it is information packet
    // if (ctrl_flags.i_frame.zero_bit == 0) {
    if (size < 2) {
        *error = HDLC_OUT_OF_DATA;
        return (NULL);
    }
    swap16_t hcs_raw;
    hcs_raw.bytes[1] = *input++;
    hcs_raw.bytes[0] = *input++;
    size -= 2;
    hcs = hcs_raw.value;
    // }

    uint8_t local_lsap=0;
    // if (ctrl_flags.i_frame.zero_bit == 0) {
    // LLC with DLMS
    if (size < 3) {
        *error = HDLC_OUT_OF_DATA;
        return (NULL);
    }
    local_lsap = parse_llc(&input,error);
    if (*error != HDLC_OK) {
        return (NULL);
    }
    size -= 3;
    // } 

    if (size <= 4) {
        // FSC 2 bytes and 1 byte MAGIC and DLMS
        *error = HDLC_OUT_OF_DATA;
        return (NULL);
    }
    
    uint8_t *dedicated_input = input;
    dlms_pdu_t *res = parse_dlms(&dedicated_input, size-3, error);
    if (res == NULL) {
        return (NULL);
    }
    
    input += 2; // FSC skip
    size -= 2; // FSC
    size -= 1; // MAGIC
    
    input += size; // skip DLMS
    
    // test magic byte
    if (*input != MAGIC) {
        *error = HDLC_WRONG_MAGIC;
        return (NULL);
    }    
    
    *error = HDLC_OK;
    return (res);
}

// .................................................................
// Specific parse of LLC header

static uint8_t parse_llc(uint8_t **input, hdlc_parse_error_t *error) {
    uint8_t *pos = *input;
    if (*pos++ != LLC_MAGIC) {
        *error = HDLC_WRONG_MISSING_LLC;
        return (0);
    }
    uint8_t res_local_lsap = *pos++;
    if (*pos++ != LLC_QUALITY) {
        *error = HDLC_WRONG_MISSING_LLC;
        return (0);
    }
    *input = pos;
    *error = HDLC_OK;
    return res_local_lsap;
}

// EOF parse_hdlc
// .................................................................
// .................................................................
