/*
 * Author: Dušan Kolář
 */

#ifndef PARSE_HDLC
#define PARSE_HDLC

typedef enum e_hdlc_parse_error {
    HDLC_OK,
    HDLC_TOO_SHORT,                 // passed buffer is shorter then smallest packet
    HDLC_WRONG_MAGIC,               // wrong magic byte 0x7E on the beginning/end
    HDLC_WRONG_MISSING_LLC,         // Expected LLC header not found
    HDLC_BASE_SEQ,                  // BASE128 sequence representing number too long    
    HDLC_OUT_OF_DATA,               // running out of the passed data length
    HDLC_NOT_MATCHING_SIZE,         // size of the past data and packet length do not match
    HDLC_NOT_AN_IFRAME,             // Only I-frames carry some information/payload that is further processed
    HDLC_DLMS_TOO_SHORT,            // DLMS pdu is too short
    HDLC_DLMS_WRONG_PDU_TYPE,       // DLMS PDU type not recognized
    HDLC_DLMS_WRONG_GET_REQ_TYPE,   // DLMS PDU get request type not recognized
    HDLC_DLMS_WRONG_DATA_TYPE,      // DLMS data type unknown
    HDLC_DLMS_WRONG_SET_REQ_TYPE,   // DLMS PDU set request type not recognized
    HDLC_DLMS_WRONG_ACT_REQ_TYPE,   // DLMS PDU action request type not recognized
    HDLC_DLMS_WRONG_GET_RESP_TYPE,  // DLMS PDU get response type not recognized
    HDLC_DLMS_WRONG_ACC_RES_CHOICE, // DLMS get response with data block access response choice not recognized
    HDLC_DLMS_WRONG_SET_RESP_TYPE,  // DLMS PDU set response type not recognized
    HDLC_DLMS_WRONG_ACT_RESP_TYPE,  // DLMS PDU action response type not recognized
    HDLC_DLMS_UNSUPPORTED_LEN,      // DLMS length encoded integer of unsupported size (just 1 and 2 bytes supported)
    HDLC_DLMS_SIZE_MISMATCH,        // expected and derived length of packet do not match
    HDLC_OUT_OF_MEMORY,             // malloc failed to return free memory thunk    
} hdlc_parse_error_t;


#include <inttypes.h>

#define MAX_LIST_LENGTH       32

typedef struct s_request {
    uint16_t                       class_id;
    uint8_t                       *instance_id;           // 6 bytes there
    uint8_t                        attribute_method_id;
    uint8_t                        data_present;          // if != 0 then access_selector, data_type and possibly octet_string_len active, for OPTIONAL data
    uint8_t                        data_hard_present;     // if != 0 && data_present==0 then data_type and possibly octet_string_len active, for HARD data
    uint8_t                        access_selector;
    uint8_t                        data_result_type;
    uint8_t                        data_type;
    uint8_t                        octet_string_len;
    uint8_t                       *octet_string_start;
} request_t;

typedef struct s_dlms_pdu {
    uint8_t                        pdu_type;
    uint8_t                        pdu_sub_type;          // request/response type, 
    uint8_t                        priority;
    uint8_t                        service_class;
    uint8_t                        invoke_id;
    request_t                      req;
    uint32_t                       block_number;
    uint8_t                       *data_block;
    uint16_t                       item_count;
    uint8_t                        date_time_present;
    uint8_t                       *date_time;            // 12 bytes there
    uint8_t                        last_block;
    uint8_t                        result_choice;
    uint8_t                        data_access_result;
    uint16_t                       str_length;
    char                          *str;
    uint16_t                       list_length;
    uint8_t                        list_data_acces_result[MAX_LIST_LENGTH];
    uint8_t                        action_response_result;
} dlms_pdu_t;


dlms_pdu_t *parse_hdlc(uint8_t *input, uint32_t size, hdlc_parse_error_t *error);

#endif
